@@ -105,7 +105,7 @@ class HeadNode(Node, ABC):
 
         # Internal RDataFrame object, useful to expose information such as
         # column names.
-        self._localdf = localdf
+        self.rdf_node = localdf
 
         # A dictionary where the keys are the IDs of the objects to live visualize
         # and the values are the corresponding callback functions 
@@ -119,8 +119,8 @@ class HeadNode(Node, ABC):
         the garbage collector, the cppyy memory regulator and the C++ object
         destructor.
         """
-        if hasattr(self, "_localdf"):
-            del self._localdf
+        if hasattr(self, "rdf_node"):
+            del self.rdf_node
 
     @property
     def npartitions(self) -> Optional[int]:
@@ -182,6 +182,28 @@ class HeadNode(Node, ABC):
     def _handle_returned_values(self, values: TaskResult) -> Iterable:
         pass
 
+    def _execute_and_retrieve_results(self, mapper, local_nodes) -> TaskResult:
+        if self.drawables_dict is not None:
+            # Prepare a dictionary with additional information for live visualization
+            drawables_info_dict = {
+                # Key: node_id
+                node.node_id: (
+                    # Tuple containing:
+                    # 1. Callback functions passed by the user
+                    self.drawables_dict[node.node_id],
+                    # 2. Index of the node in the local_nodes list
+                    i,
+                    # 3. Name of the operation associated with the node
+                    node.operation.name
+                )
+                for i, node in enumerate(local_nodes)
+                # Filter: Only include nodes requested by the user
+                if node.node_id in self.drawables_dict
+            }
+            return self.backend.ProcessAndMergeLive(self._build_ranges(), mapper, distrdf_reducer, drawables_info_dict)
+        else:
+            return self.backend.ProcessAndMerge(self._build_ranges(), mapper, distrdf_reducer)
+
     def execute_graph(self) -> None:
         """
         Executes an RDataFrame computation graph on a distributed backend.
@@ -215,36 +237,25 @@ class HeadNode(Node, ABC):
 
         computation_graph_callable = partial(ComputationGraphGenerator.trigger_computation_graph, self._generate_graph_dict())
 
+        # Accumulate all code that needs to be declared in one string
+        code_to_declare = "\n".join(self.backend.strings_to_declare.values())
+
         mapper = partial(distrdf_mapper,
                          build_rdf_from_range=self._generate_rdf_creator(),
                          computation_graph_callable=computation_graph_callable,
-                         initialization_fn=self.backend.initialization)
+                         initialization_fn=self.backend.initialization,
+                         code_to_declare=code_to_declare)
 
         # List of action nodes in the same order as values
         local_nodes = self._get_action_nodes()
 
         # Execute graph distributedly and return the aggregated results from all tasks
         # using the appropriate backend method based on whether or not live visualization is enabled
-        if self.drawables_dict is not None:
-            # Prepare a dictionary with additional information for live visualization
-            drawables_info_dict = {
-                # Key: node_id
-                node.node_id: (
-                    # Tuple containing:
-                    # 1. Callback functions passed by the user
-                    self.drawables_dict[node.node_id],
-                    # 2. Index of the node in the local_nodes list
-                    i,
-                    # 3. Name of the operation associated with the node
-                    node.operation.name
-                )
-                for i, node in enumerate(local_nodes)
-                # Filter: Only include nodes requested by the user
-                if node.node_id in self.drawables_dict
-            }
-            returned_values = self.backend.ProcessAndMergeLive(self._build_ranges(), mapper, distrdf_reducer, drawables_info_dict)
-        else:
-            returned_values = self.backend.ProcessAndMerge(self._build_ranges(), mapper, distrdf_reducer)
+        try:
+            returned_values = self._execute_and_retrieve_results(mapper, local_nodes)
+        finally:
+            # Cleanup the current execution artifacts from the caches on the workers
+            self.backend.cleanup_cache(self.exec_id)
 
         # Perform any extra checks that may be needed according to the
         # type of the head node
@@ -253,17 +264,13 @@ class HeadNode(Node, ABC):
         # Set the value of every action node
         for node, value in zip(local_nodes, final_values):
             Utils.set_value_on_node(value, node, self.backend)
-            
-    def GetColumnNames(self) -> Iterable[str]:
-        return self._localdf.GetColumnNames()
 
 
 def get_headnode(backend: BaseBackend, npartitions: int, *args) -> HeadNode:
     """
     A factory for different kinds of head nodes of the RDataFrame computation
-    graph, depending on the arguments to the RDataFrame constructor. Currently
-    can return a TreeHeadNode or an EmptySourceHeadNode. Parses the arguments and
-    compares them against the possible RDataFrame constructors.
+    graph, depending on the arguments to the RDataFrame constructor. Parses the
+    arguments and compares them against the possible RDataFrame constructors.
     """
 
     # Early check that arguments are accepted by RDataFrame
@@ -272,60 +279,23 @@ def get_headnode(backend: BaseBackend, npartitions: int, *args) -> HeadNode:
     except TypeError:
         raise TypeError(("The arguments provided are not accepted by any RDataFrame constructor. "
                          "See the RDataFrame documentation for the accepted constructor argument types."))
-        
-    firstarg = args[0]
-    if isinstance(firstarg, int):
-        # RDataFrame(ULong64_t numEntries)
-        return EmptySourceHeadNode(backend, npartitions, localdf, firstarg)
-    elif isinstance(firstarg, (ROOT.TTree)):
-    #     # RDataFrame(std::string_view treeName, filenameglob, defaultBranches = {})
-    #     # RDataFrame(std::string_view treename, filenames, defaultBranches = {})
-    #     # RDataFrame(std::string_view treeName, dirPtr, defaultBranches = {})
-    #     # RDataFrame(TTree &tree, const ColumnNames_t &defaultBranches = {})
-        return TreeHeadNode(backend, npartitions, localdf, *args) 
-    elif isinstance(firstarg, str):
-        # if first argument is a string, we want to check if the second argument
-        # which is a ROOT file holds a TTree or RNTuple
-        secondarg = args[1]
 
-        if (isinstance (secondarg, (str, ROOT.std.string, ROOT.std.string_view))):
-            wildcards = ["[", "]", "*", "?"]
-            if (any(wildcard in secondarg for wildcard in wildcards)):
-                path = ROOT.Internal.TreeUtils.ExpandGlob(secondarg)[0]
-            else:
-                path = secondarg
-        else:
-            path = secondarg[0]
+    if isinstance(args[0], ROOT.RDF.Experimental.RDatasetSpec):
+        return RDatasetSpecHeadNode(backend, npartitions, localdf, *args)
 
-        with ROOT.TDirectory.TContext(), ROOT.TFile.Open(path, "READ_WITHOUT_GLOBALREGISTRATION") as f:
-            dataset = f.Get(firstarg)
-            try:
-                if isinstance(dataset, ROOT.TTree):
-                    return TreeHeadNode(backend, npartitions, localdf, *args)
-
-                elif isinstance(dataset, ROOT.Experimental.RNTuple):
-                    return RNTupleHeadNode(backend, npartitions, localdf, *args)
-
-                else:
-                    raise RuntimeError("")
-            finally:
-                # Remove the reference to the object taken from the TFile
-                # This helps avoiding conflicts between the Python garbage
-                # collector, the cppyy memory regulator and the C++ object.
-                # TODO: Rewrite this whole section as a C++ function. Idea:
-                # the function should be a method of RInterfaceBase so that
-                # from the localdf object we could get a data source label and
-                # dispatch to the correct HeadNode type without reopening.
-                del dataset
-    elif isinstance(firstarg, ROOT.RDF.Experimental.RDatasetSpec):
-        # RDataFrame(rdatasetspec)
-        return RDatasetSpecHeadNode(backend, npartitions, localdf, *args)      
+    label = ROOT.Internal.RDF.GetDataSourceLabel(ROOT.RDF.AsRNode(localdf))
+    if label == "TTreeDS":
+        return TreeHeadNode(backend, npartitions, localdf, *args)
+    elif label == "RNTupleDS":
+        return RNTupleHeadNode(backend, npartitions, localdf, *args)
+    elif label == "EmptyDS":
+        return EmptySourceHeadNode(backend, npartitions, localdf, args[0])
     else:
         raise RuntimeError(
-            ("First argument {} of type {} is not recognised as a supported "
-                "argument for distributed RDataFrame. Currently only TTree/Tchain "
-                "based datasets or datasets created from a number of entries "
-                "can be processed distributedly.").format(firstarg, type(firstarg)))
+            (f"First argument {args[0]} of type {type(args[0])} is not "
+             "recognised as a supported argument for distributed RDataFrame. "
+             "Currently supported data sources are: TTree, RNTuple or an empty "
+             "data source."))
 
 
 class EmptySourceHeadNode(HeadNode):
@@ -389,15 +359,13 @@ class EmptySourceHeadNode(HeadNode):
             Builds an RDataFrame instance for a distributed mapper.
             """
             if current_range.exec_id not in _graph_cache._RDF_REGISTER:
-                rdf_toprocess = ROOT.RDataFrame(nentries)
-                _graph_cache._RDF_REGISTER[current_range.exec_id] = rdf_toprocess
-            else:
-                rdf_toprocess = _graph_cache._RDF_REGISTER[current_range.exec_id]
+                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(nentries)
 
             ROOT.Internal.RDF.ChangeEmptyEntryRange(
-                ROOT.RDF.AsRNode(rdf_toprocess), (current_range.start, current_range.end))
+                ROOT.RDF.AsRNode(_graph_cache._RDF_REGISTER[current_range.exec_id]),
+                (current_range.start, current_range.end))
 
-            return TaskObjects(rdf_toprocess, None)
+            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], None)
 
         return build_rdf_from_range
 
@@ -519,7 +487,7 @@ class TreeHeadNode(HeadNode):
             # Depending on the cluster setup, this may still be quite costly, so
             # we decide to pay the price only if the user explicitly requested
             # warning logging.
-            clusters, entries = Ranges.get_clusters_and_entries(self.subtreenames[0], self.inputfiles[0])
+            clusters, entries = ROOT.Internal.TreeUtils.GetClustersAndEntries(self.subtreenames[0], self.inputfiles[0])
             # The file could contain an empty tree. In that case, the estimate will not be computed.
             if entries > 0:
                 partitionsperfile = self.npartitions / len(self.inputfiles)
@@ -584,16 +552,15 @@ class TreeHeadNode(HeadNode):
             attach_friend_info_if_present(clustered_range, ds)
 
             if current_range.exec_id not in _graph_cache._RDF_REGISTER:
-                rdf_toprocess = ROOT.RDataFrame(ds)
                 # Fill the cache with the new RDataFrame
-                _graph_cache._RDF_REGISTER[current_range.exec_id] = rdf_toprocess
+                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(ds)
             else:
-                # Retrieve an already present RDataFrame from the cache
-                rdf_toprocess = _graph_cache._RDF_REGISTER[current_range.exec_id]
                 # Update it to the range of entries for this task
-                ROOT.Internal.RDF.ChangeSpec(ROOT.RDF.AsRNode(rdf_toprocess), ROOT.std.move(ds))
+                ROOT.Internal.RDF.ChangeSpec(
+                    ROOT.RDF.AsRNode(_graph_cache._RDF_REGISTER[current_range.exec_id]),
+                    ROOT.std.move(ds))
 
-            return TaskObjects(rdf_toprocess, entries_in_trees)
+            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], entries_in_trees)
 
         return build_rdf_from_range
 
@@ -700,7 +667,7 @@ class RDatasetSpecHeadNode(HeadNode):
             # Depending on the cluster setup, this may still be quite costly, so
             # we decide to pay the price only if the user explicitly requested
             # warning logging.
-            clusters, entries = Ranges.get_clusters_and_entries(
+            clusters, entries = ROOT.Internal.TreeUtils.GetClustersAndEntries(
                 self.subtreenames[0], self.inputfiles[0])
             # The file could contain an empty tree. In that case, the estimate will not be computed.
             if entries > 0:
@@ -773,17 +740,15 @@ class RDatasetSpecHeadNode(HeadNode):
             attach_friend_info_if_present(clustered_range, ds)
 
             if current_range.exec_id not in _graph_cache._RDF_REGISTER:
-                rdf_toprocess = ROOT.RDataFrame(ds)
                 # Fill the cache with the new RDataFrame
-                _graph_cache._RDF_REGISTER[current_range.exec_id] = rdf_toprocess
+                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(ds)
             else:
-                # Retrieve an already present RDataFrame from the cache
-                rdf_toprocess = _graph_cache._RDF_REGISTER[current_range.exec_id]
                 # Update it to the range of entries for this task
                 ROOT.Internal.RDF.ChangeSpec(
-                    ROOT.RDF.AsRNode(rdf_toprocess), ROOT.std.move(ds))
+                    ROOT.RDF.AsRNode( _graph_cache._RDF_REGISTER[current_range.exec_id]),
+                    ROOT.std.move(ds))
 
-            return TaskObjects(rdf_toprocess, entries_in_trees)
+            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], entries_in_trees)
 
         return build_rdf_from_range
 
@@ -876,8 +841,7 @@ class RNTupleHeadNode(HeadNode):
             if not filenames:
                 return TaskObjects(None, None)
 
-            rdf_toprocess = ROOT.RDF.Experimental.FromRNTuple(ntuplename, filenames)
-            return TaskObjects(rdf_toprocess, None)
+            return TaskObjects(ROOT.RDF.FromRNTuple(ntuplename, filenames), None)
 
         return build_rdf_from_range
 
